// dllmain.cpp : Defines the entry point for the DLL application.
#include "framework.h"
#include <strsafe.h>

extern "C" void GetUserNameExW()
{
}

extern "C" void DummyCall()
{
}

class ScopedHandle
{
    HANDLE _h;
public:
    ScopedHandle() : _h(nullptr)
    {
    }

    ScopedHandle(ScopedHandle&) = delete;

    ~ScopedHandle()
    {
        if (!invalid())
        {
            CloseHandle(_h);
            _h = nullptr;
        }
    }

    bool invalid() {
        return (_h == nullptr) || (_h == INVALID_HANDLE_VALUE);
    }

    HANDLE detach()
    {
        HANDLE ret = _h;
        _h = nullptr;
        return ret;
    }

    void set(HANDLE h)
    {
        _h = h;
    }

    HANDLE get()
    {
        return _h;
    }

    HANDLE* ptr()
    {
        return &_h;
    }
};

DWORD GetSessionId()
{
    ScopedHandle token;
    if (!OpenThreadToken(GetCurrentThread(), TOKEN_QUERY, TRUE, token.ptr()))
    {
        OutputDebugStringW(L"Error opening token\n");
        return 0;
    }

    DWORD session_id = 0;
    DWORD ret_length;
    if (!GetTokenInformation(token.get(), TokenSessionId, &session_id, sizeof(session_id), &ret_length))
    {
        OutputDebugStringW(L"Error querying token\n");
        return 0;
    }
    return session_id;
}

void StartProcess()
{
    DWORD session_id = GetSessionId();
    RevertToSelf();
 
    ScopedHandle token;
    if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ALL_ACCESS, token.ptr()))
    {
        OutputDebugStringW(L"Error opening process token\n");
        return;
    }

    ScopedHandle new_token;
    if (!DuplicateTokenEx(token.get(), TOKEN_ALL_ACCESS, nullptr, SecurityAnonymous, TokenPrimary, new_token.ptr()))
    {
        OutputDebugStringW(L"Error duplicating process token\n");
        return;
    }

    SetTokenInformation(new_token.get(), TokenSessionId, &session_id, sizeof(session_id));

    STARTUPINFO start_info = {};
    start_info.cb = sizeof(start_info);
    start_info.lpDesktop = L"WinSta0\\Default";
    PROCESS_INFORMATION proc_info;
    WCHAR cmdline[] = L"cmd.exe";
    if (CreateProcessAsUser(new_token.get(), nullptr, cmdline,
        nullptr, nullptr, FALSE, CREATE_NEW_CONSOLE, nullptr, nullptr, &start_info, &proc_info))
    {
        CloseHandle(proc_info.hProcess);
        CloseHandle(proc_info.hThread);
    }
}

BOOL APIENTRY DllMain( HMODULE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
                     )
{
    switch (ul_reason_for_call)
    {
    case DLL_PROCESS_ATTACH:
        StartProcess();
        break;
    case DLL_THREAD_ATTACH:
    case DLL_THREAD_DETACH:
    case DLL_PROCESS_DETACH:
        break;
    }
    return TRUE;
}

